import 'package:dio/dio.dart';
import 'package:flutter/material.dart';
import 'package:lori_sdk/data/lori_api.dart';
import 'package:lori_sdk/data/models/cart/seller.dart';
import 'package:lori_sdk/data/models/configurations/configurations.dart';
import 'package:lori_sdk/data/models/configurations/internalConfigurations.dart';

class SalesRetailerApi {
  static const API_RETAILER_CONFIGURATIONS = '/sales/retailer/configurations';
  static const API_RETAILER_INTERNAL_CONFIGURATIONS = '/sales/retailer/configurations/internal';

  static const API_SELLERS_FOR_RETAILER = '/sales/retailer/sellers';


  Dio _requester;

  SalesRetailerApi({@required Dio requester}) : this._requester = requester;

  Future<ApiCallback<Configurations>> loadConfigurations() async {
    try {
      var response = await _requester.get(API_RETAILER_CONFIGURATIONS);
      var configurations = Configurations.fromJson(response.data);
      // Callback
      return ApiCallback<Configurations>(
          isSuccessful: true, resultData: configurations);
    } on DioError catch (e) {
      return ApiCallback<Configurations>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<InternalConfigurations>> loadInternalConfigurations() async {
    try {
      var response = await _requester.get(API_RETAILER_INTERNAL_CONFIGURATIONS);
      var internalConfigurations = InternalConfigurations.fromJson(response.data);
      return ApiCallback<InternalConfigurations>(
          isSuccessful: true, resultData: internalConfigurations);
    } on DioError catch (e) {
      return ApiCallback<InternalConfigurations>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<Seller>>> getSellers(
      {@required double latitude, @required double longitude}) async {
    try {
      Map<String, dynamic> queryParameters = {
        'lat': latitude,
        'lng': longitude
      };
      var response = await _requester.get(API_SELLERS_FOR_RETAILER,
          queryParameters: queryParameters,
          options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var sellers = List<Seller>();
      for (var row in responseData) {
        sellers.add(Seller.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Seller>>(isSuccessful: true, resultData: sellers);
    } on DioError catch (e) {
      return ApiCallback<List<Seller>>().defaultErrorCallback(e);
    }
  }
}
