import 'dart:async';
import 'dart:convert';
import 'package:dio/dio.dart';
import 'package:firebase_performance/firebase_performance.dart';
import 'package:flutter/src/widgets/framework.dart' show BuildContext;
import 'package:flutter_phoenix/flutter_phoenix.dart';
import 'package:lori_sdk/data/api/sales/sales_api.dart';
import 'package:lori_sdk/data/api/firebase/firebase_api.dart';
import 'package:lori_sdk/data/database/interfaces/authResponseDao.dart';
import 'package:lori_sdk/data/database/interfaces/cartDao.dart';
import 'package:lori_sdk/data/database/interfaces/userDao.dart';
import 'package:lori_sdk/data/lori_push_notification_token.dart';
import 'package:lori_sdk/data/models/addressCep.dart';
import 'package:lori_sdk/data/models/addressCity.dart';
import 'package:lori_sdk/data/models/addressState.dart';
import 'package:lori_sdk/data/models/authResponse.dart';
import 'package:lori_sdk/data/models/banner.dart';
import 'package:lori_sdk/data/models/cart/cart.dart';
import 'package:lori_sdk/data/models/cart/cartItem.dart';
import 'package:lori_sdk/data/models/cart/freight.dart';
import 'package:lori_sdk/data/models/cart/installment.dart';
import 'package:lori_sdk/data/models/cart/paymentMethod.dart';
import 'package:lori_sdk/data/models/cart/seller.dart';
import 'package:lori_sdk/data/models/cart/sellerModality.dart';
import 'package:lori_sdk/data/models/check_email_response.dart';
import 'package:lori_sdk/data/models/configurations/configurations.dart';
import 'package:lori_sdk/data/models/faq.dart';
import 'package:lori_sdk/data/models/order/complaint.dart';
import 'package:lori_sdk/data/models/order/complaintType.dart';
import 'package:lori_sdk/data/models/order/order.dart';
import 'package:lori_sdk/data/models/product/evaluation.dart';
import 'package:lori_sdk/data/models/product/evaluationForm.dart';
import 'package:lori_sdk/data/models/product/opinionResponse.dart';
import 'package:lori_sdk/data/models/product/product.dart';
import 'package:lori_sdk/data/models/product/productOrder.dart';
import 'package:lori_sdk/data/models/product/searchTerm.dart';
import 'package:lori_sdk/data/models/promotion.dart';
import 'package:lori_sdk/data/models/promotionPolicy.dart';
import 'package:lori_sdk/data/models/response/productPaginationResult.dart';
import 'package:lori_sdk/data/models/response/storeResult.dart';
import 'package:lori_sdk/data/models/smsRequestResult.dart';
import 'package:lori_sdk/data/models/user.dart';
import 'package:lori_sdk/data/models/userAddress.dart';
import 'package:lori_sdk/data/models/user_promotions/userPromotion.dart';
import 'package:lori_sdk/data/models/v2/category.dart';
import 'package:lori_sdk/misc/lori_sdk_exception.dart';
import 'package:validators/validators.dart';
import 'package:flutter/foundation.dart' show kIsWeb;

import 'models/privacyPolicy.dart';
import 'models/userTerms.dart';

/// region Legacy Routes
const GRANT_TYPE_CLIENT = "client_credentials";
const GRANT_TYPE_USER = "password";
const GRANT_TYPE_SMS = "sms";
const GRANT_TYPE_REFRESH_TOKEN = "refresh_token";
const SCOPE = "lori offline_access";

const LOGIN = 'connect/token';
const CHECK_EMAIL = 'api/buyers/createatretailer';
const CHECK_CPF_PROZ = 'api/buyers/propz/validate';
const CONFIGURATIONS = 'api/configurations';
const DEVICE_TOKEN = 'api/users/devicetoken';
const NOTIFICATION_OPENED = '/api/notifications/%/opened';
const BANNERS = 'api/components/banners';
const PROMOTIONS = 'api/promotions';
const FAQ = 'api/faq';
const PRODUCT_EVALUATIONS = 'api/products/%/evaluation/responses';
const PRODUCT_EVALUATION_FORM = 'api/products/evaluation';
const PRODUCT_EVALUATION = 'api/products/%/evaluation/responses';
const PRODUCT_WISHLIST = 'api/buyers/favorites/';
const WISHLIST = 'api/buyers/favorites';
const COMPLAINTS = 'api/feedbacks';
const COMPLAINT_TYPES = 'api/feedbacks/types';

const CART = 'api/carts';
const CART_PRODUCT = 'api/carts/items';
const CART_GIFT = 'api/carts/items/gift';
const CART_OWNER = 'api/carts/buyer';
const CART_LOCATION = 'api/carts/location';
const CART_ADDRESS = 'api/carts/address/%';
const CART_CEP = 'api/carts/cep/%';
const CART_PAYMENT = 'api/carts/paymentmethod/%';
const CART_MODALITIES = 'api/carts/modalities';
const CART_MODALITY = 'api/carts/modality/%';
const CART_SELLERS = 'api/carts/sellers';
const CART_SELLERS_ITEMS = 'api/carts/items/sellers/%';
const CART_SELLER = 'api/carts/seller/%';
const CART_INSTALLMENTS = 'api/carts/installments';
const CART_INSTALLMENT = 'api/carts/installments/%';
const CART_FREIGHTS = 'api/cart/freights';
const CART_FREIGHT = 'api/carts/freight/%';
const CART_PICKUP_CPF = 'api/carts/cpfpickup/%';
const CART_INVOICE_CPF = 'api/carts/cpfinvoice/%';
const CART_CVV = 'api/carts/cvv/%';
const CART_VOUCHER = 'api/carts/voucher/%';
const DELETE_CART_VOUCHER = 'api/carts/voucher';
const CART_PAYMENTS = 'api/carts/paymentmethods';
const CART_DEALER_CODE = 'api/carts/dealercode/%';
const CART_FINGERPRINT = 'api/carts/devicefingerprint/%';
const CART_FINALIZE = 'api/carts/finalize';

const ORDERS = 'api/orders';
const ORDER_BY_ID = 'api/orders/%';
const ORDER_PAYMENTS = 'api/orders/%/paymentmethods';
const UPDATE_ORDER_PAYMENT = 'api/orders/%/paymentmethod/%';
const UPDATE_ORDER_PAYMENT_CVV = 'api/orders/%/cvv/%';
const ORDER_INSTALLMENTS = 'api/orders/%/installments';
const UPDATE_ORDER_INSTALLMENTS = 'api/orders/%/installments/%';
const RESTART_ORDER = 'api/orders/%/restart';

const PRODUCT_BY_BARCODE = "api/products/barcode/%";
const PRODUCT_BY_ID = "api/products/";
const PRODUCT_RELATED = 'api/products/%/related';
const PRODUCTS_BY_TAG = "api/products/search";
const PRODUCT_CATEGORIES = 'api/categories';
const PRODUCT_CATEGORIES_V2 = 'api/v2/categories';

@Deprecated("Category V1 has been removed")
const PRODUCT_CATEGORY = 'api/categories/%';
const PRODUCT_CATEGORY_V2 = 'api/v2/categories/%';

@Deprecated("Category V1 has been removed")
const PRODUCT_SUBCATEGORIES = 'api/categories/%/subcategories';
const PRODUCT_CHILDREN_CATEGORIES_V2 = 'api/v2/categories/%/children';
const PRODUCT_FROM_SUBCATEGORIES = 'api/categories/subcategories/%/products';
const PRODUCT_FROM_CATEGORIES = 'api/categories/%/products';
const PRODUCT_FROM_CATEGORIES_V2 = 'api/v2/categories/%/products';
const PRODUCT_TERMS = 'api/searchs/mostwanted';
const PRODUCT_BY_TERM = 'api/products/search/%';

const USER = 'api/buyers';
const USER_IMAGE = 'api/users/image';
const USER_FORGET_PASSWORD = 'api/users/rememberpassword';
const USER_PHONE_NUMBER = 'api/users/requestchangephonenumber';
const USER_SMS_CODE = 'api/users/requestsmscode';
const USER_CHANGE_PASSWORD = 'api/users/setpasswordwithsmscode';
const USER_LOGOUT = 'api/users/logout';
const USER_ADDRESSES = 'api/users/addresses';
const USER_ADDRESS = 'api/users/addresses/%';
const USER_ADDRESS_BY_CEP = 'https://api.neomode.com.br/sales/search/cep/%';
const STATES = 'api/autocomplete/states';
const CITIES = 'api/autocomplete/states/%/citiesbycode';
const USER_PAYMENTS = 'api/buyers/paymentmethods';
const USER_PAYMENT = 'api/buyers/paymentmethods/%';
const CARD_BRAND = 'api/components/util/bininformation/%/brand';

const USER_TERMS = 'api/userterms';
const PRIVACY_POLICY = 'api/privacypolicy';
const PROMOTION_POLICY = 'api/promotionpolicy';

const STORES = 'api/search/stores';

/// endregion

/// region Api Management Routes
const LIST_USER_PROMOTIONS = "marketing/buyers/promotions";
const ACTIVATE_USER_PROMOTION = "marketing/buyers/promotions/active";

class LoriApi {
  static const _API_VERSION_V1 = 1;
  static const _API_VERSION_V2 = 2;

  static String _clientId;
  static String _clientSecret;

  /// The version of the Categories API
  ///   This data should be retrieved from the API itself
  ///   and set at the SDK as soon as possible by using
  ///   LoriApi.updateCategoryApiVersion(int categoryApiVersion)
  static int _categoryApiVersion = 2;
  static String _operatingSystem;

  static String get operatingSystem {
    return _operatingSystem;
  }

  static String _operatingSystemVersion;

  ///
  /// We are currently performing an API Migration on Lori, from the legacy api to the new one.
  /// While this migration is made, both APIs will keep working at the same time.
  ///
  /// The new API will use new methods, which are organized according
  ///   to the same convention we used on the API Management.
  ///   We have a sales API and a management API and each of those APIs have child APIs
  ///     like catalog, commerce and so on.
  ///
  /// In order to keep both apis working at the same time at the app,
  ///   the legacy api methods WILL REDIRECT TO THE NEW API when set.
  ///
  ///
  /// This migration should be performed as easy as possible to the app.
  ///   Please follow the same model listed here for ALL endpoints:
  ///
  /// The legacy GetProductByID API is located at LoriApi().loadProduct()
  /// The new GetProductById API is located at LoriApi().sales.catalog.getProduct()
  ///
  /// If _useLegacyApi is TRUE, LoriApi().loadProduct() will call the LEGACY API;
  /// If _useLegacyApi is FALSE, LoriApi().loadProduct() will redirect to the NEW API;
  ///
  /// LoriApi().sales.catalog.getProduct() will ALWAYS call the NEW API;
  ///
  static bool useLegacyCatalogApi = true;
  static bool useLegacyCommerceApi = true;
  static bool useLegacyMarketingApi = true;
  static bool useLegacyRetailerApi = true;

  /// Lori App Service API URL (LEGACY)
  static String _legacyApiUrl;

  /// Lori Api Management API URL (FUTURE)
  static String _apiBaseUrl;

  // New APIs definitions
  SalesApi sales;
  FirebaseApi firebase;

  static BuildContext appContext;

  static init(String baseUrl, String clientId, String clientSecret, String operatingSystem, String operationSystemVersion, {String apiBaseUrl}) async {
    if (baseUrl == null || !isURL(baseUrl)) {
      throw LoriSdkException('The baseUrl must be a valid url.');
    } else if (clientId == null || clientId.isEmpty) {
      throw LoriSdkException('The clientId can\'t be null or an empty string.');
    } else if (clientSecret == null || clientSecret.isEmpty) {
      throw LoriSdkException('The clientSecret can\'t be null or an empty string.');
    } else if (operatingSystem == null || operatingSystem.isEmpty) {
      throw LoriSdkException('The operatingSystem can\'t be null or an empty string.');
    }
    _legacyApiUrl = baseUrl;
    _apiBaseUrl = apiBaseUrl;
    _clientId = clientId;
    _clientSecret = clientSecret;
    _operatingSystem = operatingSystem;
    _operatingSystemVersion = operationSystemVersion;
  }

  static updateCategoryApiVersion(int categoryApiVersion) {
    if (categoryApiVersion < _API_VERSION_V1 || categoryApiVersion > _API_VERSION_V2) {
      throw LoriSdkException('The categoryApiVersion must be an int between 1 and 2.');
    }
    _categoryApiVersion = categoryApiVersion;
  }

  static int get currentCategoryApiVersion {
    return _categoryApiVersion;
  }

  static updateLegacyApiUrl(String legacyApiUrl) async {
    //Check if the urls are different
    if (legacyApiUrl != null && legacyApiUrl != _legacyApiUrl) {
      //The other DAOs, like cart, can remain
      _legacyApiUrl = legacyApiUrl;
    }
  }

  // Http configuration
  Dio _legacyHttpRequester;
  Dio _apiHttpRequester;

  // DAOs
  AuthResponseDao _authResponseDao = AuthResponseDao();
  CartDao _cartDao = CartDao();
  UserDao _userDao = UserDao();

  // Current cart
  Cart authCart;

  //Array to hold the firebase metrics while they are being measure
  Map<String, HttpMetric> _firebaseHttpMetric = {};

  LoriApi() {
    configureHttpRequester();
  }

  void _restartApp() {
    if (LoriApi.appContext != null) {
      //Restart the app
      Phoenix.rebirth(LoriApi.appContext);
    }
  }

  void configureHttpRequester() async {
    if (_apiBaseUrl != null) {
      _apiHttpRequester = Dio(BaseOptions(baseUrl: _apiBaseUrl, connectTimeout: 120000, receiveTimeout: 120000));

      _apiHttpRequester.interceptors.add(InterceptorsWrapper(onRequest: (RequestOptions options) async {
        _createMetricForRequestOptions(options);

        if (options.path != LOGIN) {
          return await setAuthToken(_apiHttpRequester, options);
        }
        return options;
      }, onResponse: (Response response) async {
        _updateMetricForResponse(response);

        return response;
      }, onError: (DioError error) async {
        // Check for unauthorized responses
        if (error.response != null) {
          if (error.response.statusCode == 401) {
            // Lock things
            _apiHttpRequester.lock();
            var authClient = await _authResponseDao.get(AuthResponse.responseClient);
            var authUser = await _authResponseDao.get(AuthResponse.responseUser);
            if (authUser == null) {
              if (authClient != null) {
                // Refresh the token
                var response = await loginClient();
                if (response.isSuccessful) {
                  _apiHttpRequester.unlock();
                  // Retry the request that has failed
                  await _apiHttpRequester.request(error.response.request.path, options: error.response.request);
                } else {
                  if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
                    return error;
                  } else {
                    await _authResponseDao.deleteAll();
                    _restartApp();
                  }
                }
              }
            } else {
              // Refresh the token
              var response = await refreshLogin(authUser);
              if (response.isSuccessful) {
                _apiHttpRequester.unlock();
                // Retry the request that has failed
                await _apiHttpRequester.request(error.response.request.path, options: error.response.request);
              } else {
                if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
                  return error;
                } else {
                  await _authResponseDao.deleteAll();
                  _restartApp();
                }
              }
            }
          } else {
            return error;
          }
        } else {
          return error;
        }
        // Return here to silence the IDE error
        return error;
      }));
    }

    _legacyHttpRequester = Dio(BaseOptions(baseUrl: _legacyApiUrl, connectTimeout: 120000, receiveTimeout: 1200000));
    // Configure all output requests with existent authorization headers
    _legacyHttpRequester.interceptors.add(InterceptorsWrapper(onRequest: (RequestOptions options) async {
      _createMetricForRequestOptions(options);

      if (options.path != LOGIN) {
        return await setAuthToken(_legacyHttpRequester, options);
      }
      return options;
    }, onResponse: (Response response) async {
      _updateMetricForResponse(response);

      return response;
    }, onError: (DioError error) async {
      // Check for unauthorized responses
      if (error.response != null) {
        if (error.response.statusCode == 401) {
          // Lock things
          _legacyHttpRequester.lock();
          var authClient = await _authResponseDao.get(AuthResponse.responseClient);
          var authUser = await _authResponseDao.get(AuthResponse.responseUser);
          if (authUser == null) {
            if (authClient != null) {
              // Refresh the token
              var response = await loginClient();
              if (response.isSuccessful) {
                _legacyHttpRequester.unlock();
                // Retry the request that has failed
                await _legacyHttpRequester.request(error.response.request.path, options: error.response.request);
              } else {
                if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
                  return error;
                } else {
                  await _authResponseDao.deleteAll();
                  _restartApp();
                }
              }
            }
          } else {
            // Refresh the token
            var response = await refreshLogin(authUser);
            if (response.isSuccessful) {
              _legacyHttpRequester.unlock();
              // Retry the request that has failed
              await _legacyHttpRequester.request(error.response.request.path, options: error.response.request);
            } else {
              if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
                return error;
              } else {
                await _authResponseDao.deleteAll();
                _restartApp();
              }
            }
          }
        } else {
          return error;
        }
      } else {
        return error;
      }
      // Return here to silence the IDE error
      return error;
    }));

    // Init APIs
    this.sales = SalesApi(requester: _apiHttpRequester);
    this.firebase = FirebaseApi(requester: _apiHttpRequester);
  }

  Future<RequestOptions> setAuthToken(Dio requester, RequestOptions options) async {
    requester.lock();
    // Check for cart id header
    if (!options.headers.containsKey('cartid')) {
      authCart = await _cartDao.get();
      if (authCart != null) {
        options.headers['cartId'] = authCart.id;
        options.headers['channel'] = _operatingSystem;
      }
    }
    // Check if the token needs to be refreshed
    var authClient = await _authResponseDao.get(AuthResponse.responseClient);
    var authUser = await _authResponseDao.get(AuthResponse.responseUser);
    // Check for auth user
    if (authUser != null) {
      // Needs to be valid
      if (!authUser.isTokenValid()) {
        // If not, make sure we update it before adding as header
        var response = await refreshLogin(authUser);
        if (response.isSuccessful) {
          authUser = await _authResponseDao.get(AuthResponse.responseUser);
          // Success, release the interceptor and add the new auth as header
          options.headers['Authorization'] = 'Bearer ' + authUser.accessToken;
          requester.unlock();
          return options;
        } else {
          if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
            options.headers['Authorization'] = 'Bearer ' + authUser.accessToken;
            requester.unlock();
            return options;
          } else {
            await _authResponseDao.deleteAll();
            await _userDao.deleteAll();
            _restartApp();
          }
        }
      } else {
        options.headers['Authorization'] = 'Bearer ' + authUser.accessToken;
        requester.unlock();
        return options;
      }
    } else if (authClient != null) {
      // Needs to be valid
      if (!authClient.isTokenValid()) {
        // Refresh the token
        var response = await loginClient();
        if (response.isSuccessful) {
          var authClient = await _authResponseDao.get(AuthResponse.responseClient);
          // Success, release the interceptor and add the new auth as header
          options.headers['Authorization'] = 'Bearer ' + authClient.accessToken;
          requester.unlock();
          return options;
        } else {
          if (response.error == null || response.error.type == DioErrorType.DEFAULT) {
            options.headers['Authorization'] = 'Bearer ' + authUser.accessToken;
            requester.unlock();
            return options;
          } else {
            await _authResponseDao.deleteAll();
            _restartApp();
          }
        }
      } else {
        options.headers['Authorization'] = 'Bearer ' + authClient.accessToken;
        requester.unlock();
        return options;
      }
    } else {
      //No token created at all.
      //Create one from scratch!
      var response = await loginClient();
      if (response.isSuccessful) {
        var authClient = await _authResponseDao.get(AuthResponse.responseClient);
        // Success, release the interceptor and add the new auth as header
        options.headers['Authorization'] = 'Bearer ' + authClient.accessToken;
        requester.unlock();
        return options;
      }
    }
    requester.unlock();
  }

  ///
  /// region Firebase Performance Helper Methods
  /// [START]
  ///

  void _createMetricForRequestOptions(RequestOptions options) async {
    // Run this only if this it not running on web. This is needed for now because the firebase_performance plugin doesn't work on WEB yet.
    if (!kIsWeb) {
      String requestId = options.path;
      // Create the metric for this request
      final HttpMetric metric = FirebasePerformance.instance.newHttpMetric(options.baseUrl + options.path, _httpMethodForRequestMethod(options.method));
      // Add the metric to the global array
      _firebaseHttpMetric[requestId] = metric;
      // Start measurement
      await metric.start();
    }
  }

  void _updateMetricForResponse(Response response) async {
    // Run this only if this it not running on web. This is needed for now because the firebase_performance plugin doesn't work on WEB yet.
    if (!kIsWeb) {
      String requestId = response.request.path;
      // Retrieve the metric for this request
      HttpMetric metric = _firebaseHttpMetric.remove(requestId);
      if (metric != null) {
        // Set the data
        metric
          ..responseContentType = response.headers.value(Headers.contentTypeHeader)
          ..httpResponseCode = response.statusCode;
        // Finish measurement
        await metric.stop();
      }
    }
  }

  HttpMethod _httpMethodForRequestMethod(String requestMethod) {
    switch (requestMethod) {
      case 'GET':
        return HttpMethod.Get;
        break;
      case 'POST':
        return HttpMethod.Post;
        break;
      case 'PUT':
        return HttpMethod.Put;
        break;
      case 'PATCH':
        return HttpMethod.Patch;
        break;
      case 'DELETE':
        return HttpMethod.Delete;
        break;
      default:
        return HttpMethod.Get;
        break;
    }
  }

  ///
  /// Firebase Performance Helper Methods
  /// endregion [END]
  ///

  ///
  /// region Authentication part. Handle the authorization bearer obtainer + refresher.
  /// [START]
  ///

  Future<ApiCallback<void>> loginClient() async {
    Map<String, String> body = {
      "client_id": _clientId,
      "client_secret": _clientSecret,
      "grant_type": GRANT_TYPE_CLIENT,
      "scope": SCOPE.replaceFirst(" offline_access", ""),
    };
    // Make the request
    try {
      Response<Map<String, dynamic>> response = await Dio(BaseOptions(baseUrl: _legacyApiUrl)).post(LOGIN, data: body, options: new Options(contentType: Headers.formUrlEncodedContentType));
      // Fetch the response on go on
      var authResponse = AuthResponse.fromJson(response.data);
      authResponse.responseType = AuthResponse.responseClient;
      // Save it to the local database
      await _authResponseDao.insertOrUpdate(authResponse);
      // Run the callback
      return ApiCallback<void>(isSuccessful: true, resultData: null);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, error: e, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> loginUser({String username, String password, String smsCode}) async {
    Map<String, String> body = {
      "client_id": _clientId,
      "client_secret": _clientSecret,
      "scope": SCOPE,
    };
    if (smsCode != null) {
      body['grant_type'] = GRANT_TYPE_SMS;
      body['code'] = smsCode;
    } else if (username != null && password != null) {
      body['grant_type'] = GRANT_TYPE_USER;
      body['username'] = username;
      body['password'] = password;
    }
    // Make the request
    try {
      Response<Map<String, dynamic>> response = await _legacyHttpRequester.post(LOGIN, data: body, options: new Options(contentType: Headers.formUrlEncodedContentType));
      // Fetch the response on go on
      var authResponse = AuthResponse.fromJson(response.data);
      authResponse.responseType = AuthResponse.responseUser;
      // Save it to the local database
      await _authResponseDao.insertOrUpdate(authResponse);
      // Load user information
      var responseUser = await loadUser();
      if (responseUser.isSuccessful) {
        //Check if there is a cart created
        var cart = await CartDao().get();

        if (cart != null) {
          //If there is a cart, update the recently logged user as owner of the cart
          var responseCartOwner = await updateCartOwner();
          if (responseCartOwner.isSuccessful) {
            // Force refresh on device token
            final token = await LoriPushNotificationToken.getStoredPushNotificationToken();
            await updateNotificationToken(token);
            // Successful login the user in
            return ApiCallback<void>(isSuccessful: true);
          } else {
            // Not success, remove the _authResponse
            await _authResponseDao.delete(AuthResponse.responseUser);
            await _userDao.deleteAll();
            return ApiCallback<void>(isSuccessful: false);
          }
        } else {
          // Successful login the user in
          return ApiCallback<void>(isSuccessful: true);
        }
      } else {
        // Not success, remove user data
        await _authResponseDao.delete(AuthResponse.responseUser);
        await _userDao.deleteAll();
        return ApiCallback<void>(isSuccessful: false);
      }
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> refreshLogin(AuthResponse authResponse) async {
    Map<String, String> body = {
      "client_id": _clientId,
      "client_secret": _clientSecret,
      "grant_type": GRANT_TYPE_REFRESH_TOKEN,
      "refresh_token": authResponse.refreshToken,
    };
    try {
      var response = await Dio(BaseOptions(baseUrl: _legacyApiUrl)).post(LOGIN, data: body, options: new Options(contentType: Headers.formUrlEncodedContentType));
      // Parse the result and save it locally
      var parsedAuthResponse = AuthResponse.fromJson(response.data);
      parsedAuthResponse.responseType = AuthResponse.responseUser;
      await _authResponseDao.insertOrUpdate(parsedAuthResponse);
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, error: e, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<Configurations>> loadConfigurations() async {
    if (!LoriApi.useLegacyRetailerApi) return this.sales.retailer.loadConfigurations();

    try {
      var response = await _legacyHttpRequester.get(CONFIGURATIONS);
      var configurations = Configurations.fromJson(response.data);
      // Callback
      return ApiCallback(isSuccessful: true, resultData: configurations);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<CheckEmailResponse>> checkEmail(String email) async {
    Map<String, String> body = {
      "email": email,
    };
    try {
      var response = await _legacyHttpRequester.post(
        CHECK_EMAIL,
        data: body,
      );

      if (response.statusCode == 200) {
        return ApiCallback<CheckEmailResponse>(isSuccessful: true, resultData: CheckEmailResponse.exists);
      } else if (response.statusCode == 201) {
        return ApiCallback<CheckEmailResponse>(isSuccessful: true, resultData: CheckEmailResponse.created);
      } else {
        return ApiCallback<CheckEmailResponse>(isSuccessful: true, resultData: CheckEmailResponse.notFound);
      }
    } on DioError catch (e) {
      if (e.response.statusCode == 404) {
        //404 is not a failure, is an expected response when the user is not registered
        return ApiCallback(isSuccessful: true, resultData: CheckEmailResponse.notFound);
      } else {
        return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
      }
    }
  }

  // This function verify if the user has a login in the propz sistem and valide if the user has
  // name and phone number, case if is true, the api will the register and return the object client
  // to redirect to login, if false, the user will redirect to register
  Future<ApiCallback<Map<String, dynamic>>> checkCPFPropz(String email, String cpf) async {
    Map<String, String> body = {"email": email, "cpf": cpf, "clientId": "atakarejo"};
    try {
      var response = await _legacyHttpRequester.post(
        CHECK_CPF_PROZ,
        data: body,
      );
      if (response.statusCode == 200) {
        bool find = response.data['find'];
        User user = User.fromJson(response.data['object']);
        final Map<String, dynamic> res = {"find": find, "user": user};

        return ApiCallback<Map<String, dynamic>>(isSuccessful: true, resultData: res);
      }
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// Authentication part. Handle the authorization bearer obtainer + refresher.
  /// endregion [END]
  ///

  ///
  /// region Product part. All methods that are related with product fetching.
  /// [START]
  ///

  @Deprecated('The new API is available directly at LoriApi().sales.catalog.getProduct(String productId)')
  Future<ApiCallback<Product>> loadProduct(String productId, {bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi) return this.sales.catalog.getProduct(productId, perSeller: perSeller);

    try {
      var response = await _legacyHttpRequester.get(PRODUCT_BY_ID + productId);
      return ApiCallback<Product>(isSuccessful: true, resultData: Product.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<Product>> loadProductFromBarcode(String barcode, {bool perSeller = false}) async {
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_BY_BARCODE.replaceAll('%', barcode));
      return ApiCallback<Product>(isSuccessful: true, resultData: Product.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

//

  Future<ApiCallback<List<Product>>> loadProductRelated(String productId) async {
    var url = PRODUCT_RELATED.replaceAll('%', productId);
    try {
      var response = await _legacyHttpRequester.get(url);
      var responseData = response.data as List;
      var products = List<Product>();
      for (var row in responseData) {
        products.add(Product.fromJson(row));
      }
      return ApiCallback<List<Product>>(isSuccessful: true, resultData: products);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  @Deprecated('The new API is available directly  at LoriApi().sales.catalog.getProductsByTag')
  Future<ApiCallback<ProductPaginationResult>> loadProductsByTag(String tags, int page, int perPage, {ProductOrder order, String continuationToken, bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi) return this.sales.catalog.getProductsByTag(tag: tags, page: page, perPage: perPage, sort: order, continuationToken: continuationToken, perSeller: perSeller);

    Map<String, dynamic> queryParameters = {'tags': tags, 'page': page, 'perPage': perPage};
    if (order != null && order.parameterName != null) {
      queryParameters['order'] = order.parameterName;
    }
    try {
      var response = await _legacyHttpRequester.get(PRODUCTS_BY_TAG, queryParameters: queryParameters);
      var productPaginationResult = ProductPaginationResult.fromJson(response.data);
      return ApiCallback(isSuccessful: true, resultData: productPaginationResult);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  @Deprecated('The new API is available directly at LoriApi().sales.catalog.getCategories()')
  Future<ApiCallback<List<Category>>> loadProductCategories({bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi) return this.sales.catalog.getCategories(perSeller: perSeller);

    try {
//      if (_categoryApiVersion == _API_VERSION_V2) {
      var response = await _legacyHttpRequester.get(PRODUCT_CATEGORIES_V2);
      var responseData = response.data as List;
      var categories = List<Category>();
      for (var row in responseData) {
        categories.add(Category.fromJson(row));
      }
      categories.sort((c1, c2) => c1.name.compareTo(c2.name));
      return ApiCallback<List<Category>>(isSuccessful: true, resultData: categories);
//      }
    } catch (e) {
      // return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
      print(e);
      return ApiCallback(
        isSuccessful: false,
      );
    }
  }

  Future<ApiCallback<Category>> loadProductCategory(String categoryId) async {
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_CATEGORY_V2.replaceAll("%", categoryId));
//      if (_categoryApiVersion == _API_VERSION_V2) {
      return ApiCallback<Category>(isSuccessful: true, resultData: Category.fromJson(response.data));
//      }
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  @Deprecated('The new API is available directly at LoriApi().sales.catalog.getChildrenCategories(String parentCategoryId)')
  Future<ApiCallback<List<Category>>> loadChildrenCategories(String categoryId, {bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi) return this.sales.catalog.getChildrenCategories(parentCategoryId: categoryId, perSeller: perSeller);

    try {
      var response = await _legacyHttpRequester.get(PRODUCT_CHILDREN_CATEGORIES_V2.replaceAll("%", categoryId));
      var responseData = response.data as List;
      var categories = List<Category>();
      for (var row in responseData) {
        categories.add(Category.fromJson(row));
      }
      return ApiCallback<List<Category>>(isSuccessful: true, resultData: categories);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  @Deprecated('Subcategories v1 has been removed! Please use loadChildrenCategories(String categoryId)')
  Future<ApiCallback<List<Category>>> loadProductSubcategories(String categoryId) async {
    throw ("Subcategories v1 has been removed! Please use loadChildrenCategories(String categoryId)");
  }

  @Deprecated('Subcategories v1 has been removed! Please use loadProductsFromCategory(String categoryId)')
  Future<ApiCallback<ProductPaginationResult>> loadProductsFromSubcategory(String subcategoryId, int page, int perPage, {ProductOrder order, String continuationToken, bool perSeller = false}) async {
    return loadProductsFromCategory(subcategoryId, page, perPage, order: order, continuationToken: continuationToken, perSeller: perSeller);
  }

  @Deprecated('The new API is available directly at LoriApi().sales.catalog.getProductsByCategory')
  Future<ApiCallback<ProductPaginationResult>> loadProductsFromCategory(String categoryId, int page, int perPage, {ProductOrder order, String continuationToken, bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi)
      return this.sales.catalog.getProductsByCategory(categoryId: categoryId, page: page, perPage: perPage, sort: order, continuationToken: continuationToken, perSeller: perSeller);

    Map<String, dynamic> queryParameters = {'page': page, 'perPage': perPage};
    if (order != null && order.parameterName != null) {
      queryParameters['order'] = order.parameterName;
    }
    var url = (_categoryApiVersion == _API_VERSION_V1 ? PRODUCT_FROM_CATEGORIES : PRODUCT_FROM_CATEGORIES_V2);
    try {
      var response = await _legacyHttpRequester.get(url.replaceAll('%', categoryId), queryParameters: queryParameters);
      if (_categoryApiVersion == _API_VERSION_V1) {
        var productPaginationResult = ProductPaginationResult.fromJson(response.data);
        return ApiCallback(isSuccessful: true, resultData: productPaginationResult);
      } else {
        var responseData = response.data as List;
        var products = List<Product>();
        for (var row in responseData) {
          products.add(Product.fromJson(row));
        }
        var productPaginationResult = ProductPaginationResult();
        productPaginationResult.products = products;
        return ApiCallback(isSuccessful: true, resultData: productPaginationResult);
      }
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<String>>> loadProductTerms() async {
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_TERMS);
      var responseData = response.data as List;
      var terms = List<String>();
      for (var row in responseData) {
        terms.add(SearchTerm.fromJson(row).term);
      }
      return ApiCallback<List<String>>(isSuccessful: true, resultData: terms);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  @Deprecated('The new API is available directly  at LoriApi().sales.catalog.getProductsBySearch')
  Future<ApiCallback<ProductPaginationResult>> loadProductsByTerm(String term, int page, int perPage, {ProductOrder order, String continuationToken, bool perSeller = false}) async {
    if (!LoriApi.useLegacyCatalogApi) return this.sales.catalog.getProductsBySearch(term: term, page: page, perPage: perPage, sort: order, continuationToken: continuationToken, perSeller: perSeller);

    Map<String, dynamic> queryParameters = {'page': page, 'perPage': perPage};
    if (order != null && order.parameterName != null) {
      queryParameters['order'] = order.parameterName;
    }
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_BY_TERM.replaceAll("%", term), queryParameters: queryParameters, options: Options(responseType: ResponseType.json));
      return ApiCallback(isSuccessful: true, resultData: ProductPaginationResult.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, error: e, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// Product part. All methods that are related with product fetching.
  /// endregion [END]
  ///

  ///
  /// region CRM part. All methods that are related with customer relationship.
  /// [START]
  ///

  Future<ApiCallback<List<Banner>>> loadBanners({BannerType type = BannerType.carousel}) async {
    if (!LoriApi.useLegacyMarketingApi) return this.sales.marketing.getBanners(type: type);

    try {
      Map<String, dynamic> queryParameters = {'type': type.index};
      var response = await _legacyHttpRequester.get(BANNERS, queryParameters: queryParameters, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var banners = List<Banner>();
      for (var row in responseData) {
        banners.add(Banner.fromJson(row));
      }
      return ApiCallback<List<Banner>>(isSuccessful: true, resultData: banners);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Promotion>>> loadPromotions() async {
    try {
      var response = await _legacyHttpRequester.get(PROMOTIONS, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var promotions = List<Promotion>();
      for (var row in responseData) {
        promotions.add(Promotion.fromJson(row));
      }
      return ApiCallback<List<Promotion>>(isSuccessful: true, resultData: promotions);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Faq>>> loadFaq() async {
    try {
      var response = await _legacyHttpRequester.get(FAQ, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var faqList = List<Faq>();
      for (var row in responseData) {
        faqList.add(Faq.fromJson(row));
      }
      return ApiCallback<List<Faq>>(isSuccessful: true, resultData: faqList);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<Evaluation>> loadProductEvaluations(String productId) async {
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_EVALUATIONS.replaceAll('%', productId), options: Options(responseType: ResponseType.json));
      return ApiCallback<Evaluation>(isSuccessful: true, resultData: Evaluation.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<EvaluationForm>> loadProductEvaluationForm() async {
    try {
      var response = await _legacyHttpRequester.get(PRODUCT_EVALUATION_FORM, options: Options(responseType: ResponseType.json));
      return ApiCallback<EvaluationForm>(isSuccessful: true, resultData: EvaluationForm.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> crateProductEvaluation(String productId, List<OpinionResponse> responses) async {
    try {
      var responsesAsJson = json.encode(responses);
      await _legacyHttpRequester.post(PRODUCT_EVALUATION.replaceAll("%", productId), data: responsesAsJson, options: Options(responseType: ResponseType.json));
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Product>>> loadWishlist() async {
    if (!LoriApi.useLegacyMarketingApi) return this.sales.marketing.getFavoriteProducts();

    try {
      var response = await _legacyHttpRequester.get(WISHLIST, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var wishlist = List<Product>();
      for (var row in responseData) {
        wishlist.add(Product.fromJson(row));
      }
      return ApiCallback<List<Product>>(isSuccessful: true, resultData: wishlist);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> addProductWishlist(String productId) async {
    try {
      await _legacyHttpRequester.post(PRODUCT_WISHLIST + productId, options: Options(responseType: ResponseType.json));
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> removeProductWishlist(String productId) async {
    try {
      await _legacyHttpRequester.delete(PRODUCT_WISHLIST + productId, options: Options(responseType: ResponseType.json));
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<StoreResult>> loadStores({double latitude, double longitude, int page = 1, int perPage = 100000}) async {
    var url = STORES + "?types=Department&page=$page&perPage=$perPage";
    if (latitude != null && longitude != null) url += '&latitude=$latitude&longitude=$longitude';
    try {
      var response = await _legacyHttpRequester.get(url, options: Options(responseType: ResponseType.json));
      return ApiCallback<StoreResult>(isSuccessful: true, resultData: StoreResult.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Complaint>>> loadComplaints() async {
    try {
      var response = await _legacyHttpRequester.get(COMPLAINTS, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var complaints = List<Complaint>();
      for (var row in responseData) {
        complaints.add(Complaint.fromJson(row));
      }
      return ApiCallback<List<Complaint>>(isSuccessful: true, resultData: complaints);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<ComplaintType>>> loadComplaintTypes() async {
    try {
      var response = await _legacyHttpRequester.get(COMPLAINT_TYPES, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var complaintTypes = List<ComplaintType>();
      for (var row in responseData) {
        complaintTypes.add(ComplaintType.fromJson(row));
      }
      return ApiCallback<List<ComplaintType>>(isSuccessful: true, resultData: complaintTypes);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> createComplaint(String description, String orderId, String complaintType, {String externalOrderId}) async {
    Map body = {'message': description, 'orderId': orderId, 'isComplaint': true.toString(), 'type': complaintType};
    if (externalOrderId != null) body['externalOrderId'] = externalOrderId;
    try {
      await _legacyHttpRequester.post(COMPLAINTS, data: body, options: Options(responseType: ResponseType.json));
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<UserTerms>> loadUserTerms() async {
    try {
      var response = await _legacyHttpRequester.get(USER_TERMS, options: Options(responseType: ResponseType.json));
      return ApiCallback<UserTerms>(isSuccessful: true, resultData: UserTerms.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<PrivacyPolicy>> loadPrivacyPolicy() async {
    try {
      var response = await _legacyHttpRequester.get(PRIVACY_POLICY, options: Options(responseType: ResponseType.json));
      return ApiCallback<PrivacyPolicy>(isSuccessful: true, resultData: PrivacyPolicy.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<PromotionPolicy>> loadPromotionPolicy() async {
    try {
      var response = await _legacyHttpRequester.get(PROMOTION_POLICY, options: Options(responseType: ResponseType.json));
      return ApiCallback<PromotionPolicy>(isSuccessful: true, resultData: PromotionPolicy.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// CRM part. All methods that are related with customer relationship.
  /// endregion [END]
  ///

  ///
  /// region Cart part. All methods that are related with app's cart.
  /// [START]
  ///

  Future<ApiCallback<void>> validateCart() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.validateCart();
    Cart cart = await _cartDao.get();
    // Check if we already have one. If yes just say that our cart is validated
    if (cart != null) {
      return ApiCallback(isSuccessful: true);
    } else {
      // If no, make sure to create one
      ApiCallback createCartResult = await createCart();
      if (createCartResult.isSuccessful) {
        return ApiCallback(isSuccessful: true);
      } else {
        // We can't continue if we don't succeed on creating a cart
        return ApiCallback(isSuccessful: false);
      }
    }
  }

  Future<ApiCallback<Cart>> createCart({String channel, bool saveAsDefault = true}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getCart(saveAsDefault: saveAsDefault, channel: channel);

    try {
      Map<String, String> body = {"latitude": '0.000000', "longitude": '0.000000', "channel": channel ?? _operatingSystem};
      var response = await _legacyHttpRequester.post(CART, data: body, options: Options(responseType: ResponseType.json));
      // Parse the cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally, if set to be default
      if (saveAsDefault) {
        await _cartDao.insertOrUpdate(parsedCart);
        // Store it to be used later
        authCart = parsedCart;
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true, resultData: parsedCart);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<Cart>> loadCart({String cartId, bool saveAsDefault = true, bool reload = false, bool forceUpdate = false}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getCart(cartId: cartId, saveAsDefault: saveAsDefault);

    try {
      var headers = Map<String, dynamic>();
      if (cartId != null) {
        headers['cartid'] = cartId;
      }
      Map<String, dynamic> queryParameters = {'reload': reload};
      var options = Options(responseType: ResponseType.json, headers: headers);
      var response = await _legacyHttpRequester.get(CART, queryParameters: queryParameters, options: options);
      // Parse the cart
      var parsedCart = Cart.fromJson(response.data);
      // Check if we have a current cart
      var currentCart = await _cartDao.get();
      // If we have a current cart, update the data based on the current existing cart
      if (currentCart != null) parsedCart.isUpdatingFrom(currentCart, forceUpdate: reload || forceUpdate);
      // Save it locally, if needed
      if (saveAsDefault) {
        await _cartDao.insertOrUpdate(parsedCart);
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true, resultData: parsedCart);
    } on DioError catch (e) {
      rethrow;
//      if (e.type == DioErrorType.RESPONSE) {
//        // If an error occurred when loading the cart
//        //  we must recreate it
//        return await regenerateCart();
//      }
//      else {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
//      }
    }
  }

  Future<ApiCallback<Cart>> regenerateCart() async {
    try {
      // Remove the current cart
      _cartDao.deleteAll();

      await createCart();

      // Although the operation was successful we return false here
      //  so an alert can be displayed to the user when this happens
      return ApiCallback(isSuccessful: false, resultErrorMessage: "Atenção! Encontramos um problema no seu carrinho e todos os produtos foram removidos.\n\nPor favor adicione os itens novamente!");
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> clearCart() async {
    try {
      await _legacyHttpRequester.delete(CART, options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartItem(String skuId, int quantity, {String cartId, bool saveAsDefault = true, bool hasReloadedCart = false}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.upgradeItem(skuId: skuId, quantity: quantity, cartId: cartId, saveAsDefault: saveAsDefault);

    try {
      var headers = Map<String, dynamic>();
      if (cartId != null) {
        headers['cartid'] = cartId;
      }
      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };
      var options = Options(responseType: ResponseType.json, headers: headers);
      // Update the item
      await _legacyHttpRequester.post(CART_PRODUCT, queryParameters: queryParameters, options: options);
      if (saveAsDefault) {
        // Update the cart locally
        await loadCart(forceUpdate: true);
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      await loadCart(forceUpdate: true);
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> downgradeCartItem(String skuId, int quantity, {String cartId, bool saveAsDefault = true}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.downgradeItem(skuId: skuId, quantity: quantity, cartId: cartId, saveAsDefault: saveAsDefault);

    try {
      var headers = Map<String, dynamic>();
      if (cartId != null) {
        headers['cartid'] = cartId;
      }
      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };
      var options = Options(responseType: ResponseType.json);
      // Update the item
      await _legacyHttpRequester.delete(CART_PRODUCT, queryParameters: queryParameters, options: options);
      if (saveAsDefault) {
        // Update the cart locally
        await loadCart(forceUpdate: true);
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      if (e.type == DioErrorType.RESPONSE) {
        // Retrieve the current cart
        var currentCart = await _cartDao.get();

        // Retrieve the cart item we were attempting to remove
        var cartItem = currentCart.cartItems.firstWhere((i) => i.sku.id == skuId);

        // Regenerate the cart
        var regenerateCartResponse = await regenerateCart();

        if (currentCart.allItems.length == 1 && cartItem.quantity == quantity) {
          // If we were attempting to totally remove this item from the cart
          //  and it was the only item on the cart a successful request would
          //  have resulted in an empty cart.
          // So, a cart recreation will be the same as a successful request.
          return ApiCallback(isSuccessful: true);
        } else {
          return regenerateCartResponse;
        }
      } else {
        return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
      }
    }
  }

  Future<ApiCallback<void>> updateCartGiftItem(String skuId, int quantity, {String cartId, bool saveAsDefault = true}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.upgradeItem(skuId: skuId, quantity: quantity, isGift: true, cartId: cartId, saveAsDefault: saveAsDefault);

    try {
      var headers = Map<String, dynamic>();
      if (cartId != null) {
        headers['cartid'] = cartId;
      }
      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };
      var options = Options(responseType: ResponseType.json, headers: headers);
      // Update the item
      await _legacyHttpRequester.post(CART_GIFT, queryParameters: queryParameters, options: options);
      if (saveAsDefault) {
        // Update the cart locally
        await loadCart();
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, error: e, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> downgradeCartGiftItem(String skuId, int quantity, {String cartId, bool saveAsDefault = true}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.downgradeItem(skuId: skuId, quantity: quantity, isGift: true, cartId: cartId, saveAsDefault: saveAsDefault);

    try {
      var headers = Map<String, dynamic>();
      if (cartId != null) {
        headers['cartid'] = cartId;
      }
      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };
      var options = Options(responseType: ResponseType.json);
      // Update the item
      await _legacyHttpRequester.delete(CART_GIFT, queryParameters: queryParameters, options: options);
      if (saveAsDefault) {
        // Update the cart locally
        await loadCart();
      }
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, error: e, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartOwner() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putBuyer();

    try {
      // Update the item
      await _legacyHttpRequester.put(CART_OWNER, options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartLocation(String lat, String lon) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putLocation(latitude: lat, longitude: lon);

    try {
      Map params = {'latitude': lat, 'longitude': lon};
      // Update cart location
      await _legacyHttpRequester.put(CART_LOCATION, data: params, options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<SellerModality>>> loadCartModalities({String latitude, String longitude}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getModalities(latitude: latitude, longitude: longitude);

    try {
      final response = await _legacyHttpRequester.get(CART_MODALITIES, options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var modalities = List<SellerModality>();
      for (var row in responseData) {
        modalities.add(SellerModality.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<SellerModality>>(isSuccessful: true, resultData: modalities);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartModality(String modalityId) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putModality(modalityId: modalityId);

    try {
      await _legacyHttpRequester.put(CART_MODALITY.replaceAll("%", modalityId), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Seller>>> loadCartSellers({String latitude, String longitude}) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getSellers(latitude: latitude, longitude: longitude);

    try {
      final response = await _legacyHttpRequester.get(CART_SELLERS, options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var sellers = List<Seller>();
      for (var row in responseData) {
        sellers.add(Seller.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Seller>>(isSuccessful: true, resultData: sellers);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<CartItem>>> loadCartSellerItems(String sellerId) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getSellerItems(sellerId: sellerId);

    try {
      final response = await _legacyHttpRequester.get(CART_SELLERS_ITEMS.replaceAll("%", sellerId), options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var cartItems = List<CartItem>();
      for (var row in responseData) {
        cartItems.add(CartItem.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<CartItem>>(isSuccessful: true, resultData: cartItems);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartSeller(String sellerId) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putSeller(sellerId: sellerId);

    try {
      await _legacyHttpRequester.put(CART_SELLER.replaceAll("%", sellerId), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart(forceUpdate: true, reload: true);
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      await loadCart(forceUpdate: true);
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartAddress(String addressId) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putAddress(addressId: addressId);

    try {
      await _legacyHttpRequester.put(CART_ADDRESS.replaceAll("%", addressId), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartCep(String cep) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putCep(cep: cep);

    try {
      await _legacyHttpRequester.put(CART_CEP.replaceAll("%", cep), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Freight>>> loadCartFreights() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getFreights();

    try {
      final response = await _legacyHttpRequester.get(CART_FREIGHTS, options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var freights = List<Freight>();
      for (var row in responseData) {
        freights.add(Freight.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Freight>>(isSuccessful: true, resultData: freights);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartFreight(String freightReference) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putFreight(freightReference: freightReference);

    try {
      await _legacyHttpRequester.put(CART_FREIGHT.replaceAll("%", freightReference), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Installment>>> loadCartInstallments() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getInstallments();

    try {
      final response = await _legacyHttpRequester.get(CART_INSTALLMENTS, options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var installments = List<Installment>();
      for (var row in responseData) {
        installments.add(Installment.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Installment>>(isSuccessful: true, resultData: installments);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartInstallment(String installmentNumber) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putInstallments(installments: installmentNumber);

    try {
      await _legacyHttpRequester.put(CART_INSTALLMENT.replaceAll("%", installmentNumber), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartPickupCpf(String cpf) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putCPFPickup(cpf: cpf);

    try {
      await _legacyHttpRequester.put(CART_PICKUP_CPF.replaceAll("%", cpf), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<PaymentMethod>>> loadCartPayments() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.getPaymentMethods();

    try {
      var response = await _legacyHttpRequester.get(CART_PAYMENTS, options: Options(responseType: ResponseType.json));
      // Fetch payment methods
      var responseData = response.data as List;
      var payments = List<PaymentMethod>();
      for (var row in responseData) payments.add(PaymentMethod.fromJson(row));
      // Callback invocation
      return ApiCallback<List<PaymentMethod>>(isSuccessful: true, resultData: payments);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartPaymentMethod(String paymentId) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putPaymentMethod(paymentMethodId: paymentId);

    try {
      await _legacyHttpRequester.put(CART_PAYMENT.replaceAll("%", paymentId), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartVoucher(String voucherCode) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putVoucher(voucherCode: voucherCode);

    try {
      await _legacyHttpRequester.put(CART_VOUCHER.replaceAll("%", Uri.encodeQueryComponent(voucherCode)), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> removeCartVoucher(String voucherCode) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.deleteVoucher();

    try {
      await _legacyHttpRequester.delete(DELETE_CART_VOUCHER, options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartCvv(String cvv) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putCvv(cvv: cvv);

    try {
      await _legacyHttpRequester.put(CART_CVV.replaceAll("%", cvv), options: Options(responseType: ResponseType.json));
      // Update the cart locally
      await loadCart();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> finalizeCart() async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.finalize();

    try {
      await _legacyHttpRequester.put(CART_FINALIZE, options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartDealerCode(String dealerCode) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putDealerCode(dealerCode: dealerCode);

    try {
      await _legacyHttpRequester.put(CART_DEALER_CODE.replaceAll("%", dealerCode), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateCartFingerprint(String deviceFingerprint) async {
    if (!LoriApi.useLegacyCommerceApi) return this.sales.commerce.carts.putDeviceFingerprint(fingerprintToken: deviceFingerprint);

    try {
      await _legacyHttpRequester.put(CART_FINGERPRINT.replaceAll("%", deviceFingerprint), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// Cart part. All methods that are related with app's cart.
  /// endregion [END]
  ///

  ///
  /// region User part. All methods that are related with user.
  /// [START]
  ///

  Future<ApiCallback<AddressCep>> loadAddressByCep(String cep) async {
    return this.sales.search.getAddressByCep(cep);

    // try {
    //   // Update the item
    //   var response = await Dio().get(
    //       USER_ADDRESS_BY_CEP.replaceAll("%", cep),
    //       options: Options(responseType: ResponseType.json));
    //   // Callback invocation
    //   return ApiCallback<AddressCep>(
    //       isSuccessful: true, resultData: AddressCep.fromJson(response.data));
    // } on DioError catch (e) {
    //   return ApiCallback(
    //       isSuccessful: false,
    //       resultErrorMessage:
    //           parseResponseError(e.response != null ? e.response.data : null));
    // }
  }

  Future<ApiCallback<List<AddressState>>> loadStates() async {
    try {
      // Update the item
      var response = await _legacyHttpRequester.get(STATES, options: Options(responseType: ResponseType.json));
      // Parse items
      var responseData = response.data as List;
      var states = List<AddressState>();
      for (var row in responseData) {
        states.add(AddressState.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<AddressState>>(isSuccessful: true, resultData: states);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<AddressCity>>> loadCities(String state) async {
    try {
      // Update the item
      var response = await _legacyHttpRequester.get(CITIES.replaceAll("%", state), options: Options(responseType: ResponseType.json));
      // Parse items
      var responseData = response.data as List;
      var cities = List<AddressCity>();
      for (var row in responseData) {
        cities.add(AddressCity.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<AddressCity>>(isSuccessful: true, resultData: cities);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateNotificationToken(String token) async {
    Map body = {'operationSystem': _operatingSystem, 'token': token, 'version': _operatingSystemVersion};
    try {
      // Update the item
      await _legacyHttpRequester.put(DEVICE_TOKEN, data: body, options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> notificationOpened(String notificationId) async {
    try {
      // Set notification opened
      await _legacyHttpRequester.put(NOTIFICATION_OPENED.replaceAll("%", notificationId), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> recoverUser(String email) async {
    try {
      // Update the item
      await _legacyHttpRequester.post(USER_FORGET_PASSWORD, data: {'email': email}, options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> recoverPhone(String email) async {
    try {
      // Update the item
      await _legacyHttpRequester.post(USER_PHONE_NUMBER, data: {'email': email}, options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<SmsRequestResult>> requestSmsCode(String email) async {
    try {
      // Update the item
      var response = await _legacyHttpRequester.post(USER_SMS_CODE, data: {'email': email}, options: Options(responseType: ResponseType.json));
      // Parse the result
      var parsedData = SmsRequestResult.fromJson(response.data);
      // Callback invocation
      return ApiCallback<SmsRequestResult>(isSuccessful: true, resultData: parsedData);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> requestPasswordChange(String email, String smsCode, String password) async {
    try {
      // Update the item
      await _legacyHttpRequester.post(USER_CHANGE_PASSWORD, data: {'email': email, 'code': smsCode, 'password': password}, options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback<SmsRequestResult>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<User>> createUser(String email, String fullName, String cpf, String phoneNumber, String password, {String dateOfBirth, bool acceptedPromotionPolicy}) async {
    try {
      Map body = {'email': email, 'fullName': fullName, 'cpf': cpf, 'phoneNumber': phoneNumber, 'password': password, 'acceptedUserTerms': true};
      if (dateOfBirth != null) {
        body['dateOfBirth'] = dateOfBirth;
      }
      if (acceptedPromotionPolicy != null) {
        body['acceptedPromotionPolicy'] = acceptedPromotionPolicy;
      }
      // Update the item
      var response = await _legacyHttpRequester.post(USER, data: body, options: Options(responseType: ResponseType.json));
      // Parse the user
      var parsedUser = User.fromJson(response.data);
      // Callback invocation
      return ApiCallback<User>(isSuccessful: true, resultData: parsedUser);
    } on DioError catch (e) {
      return ApiCallback<User>(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<User>> loadUser() async {
    try {
      var response = await _legacyHttpRequester.get(USER, options: Options(responseType: ResponseType.json));
      // Parse the user
      User parsedUser = User.fromJson(response.data);
      parsedUser.addresses = parsedUser.addresses.where((address) => !address.isDeleted).toList();
      parsedUser.paymentMethods.clear();
      // Check if we have a user already saved
      var currentUser = await _userDao.get();
      if (currentUser != null) {
        // We have a user already saved
        if (currentUser.registration != null) {
          // Keep the registration value
          parsedUser.registration = currentUser.registration;
        }
      }
      // Save it locally
      await _userDao.insertOrUpdate(parsedUser);
      // Load payment methods
      await loadUserPayments();
      // Callback invocation
      return ApiCallback<User>(isSuccessful: true, resultData: parsedUser);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateUser(String fullName, String cpf, String phoneNumber, {String dateOfBirth, int gender, String password}) async {
    try {
      Map body = {
        'fulName': fullName,
        'cpf': cpf,
        'phoneNumber': phoneNumber,
      };
      if (dateOfBirth != null) {
        body['dateOfBirth'] = dateOfBirth;
      }
      if (gender != null) {
        body['gender'] = gender;
      }
      if (password != null) {
        body['password'] = password;
      }
      await _legacyHttpRequester.patch(USER, data: body, options: Options(responseType: ResponseType.json));
      // Parse the user
      User loggedUser = await _userDao.get();
      // Update fields
      loggedUser.fullName = fullName;
      loggedUser.cpf = cpf;
      loggedUser.phoneNumber = phoneNumber;
      loggedUser.dateOfBirth = dateOfBirth != null ? dateOfBirth : loggedUser.dateOfBirth;
      loggedUser.gender = gender != null ? gender : loggedUser.gender;
      await _userDao.insertOrUpdate(loggedUser);
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<String>> updateUserPhoto(List<int> bytes) async {
    // MultipartFile.fromBytes(bytes, filename: 'imageFile.jpg',);
    // ContentType('image', 'jpeg')
    FormData formData = FormData.fromMap({
      "imageFile": MultipartFile.fromBytes(
        bytes,
        filename: 'imageFile.jpg',
      ),
    });
    try {
      await _legacyHttpRequester.patch(USER_IMAGE, data: formData, options: Options(contentType: 'image/jpeg'));
      // Update fields
      await loadUser();
      // Get user
      final loggedUser = await _userDao.get();
      // Callback invocation
      return ApiCallback<String>(isSuccessful: true, resultData: loggedUser.imageUrl);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<UserAddress>> createUserAddress(UserAddress body) async {
    Map jsonBody = body.toJson();
    jsonBody.remove('id');
    try {
      var response = await _legacyHttpRequester.post(USER_ADDRESSES, data: jsonBody, options: Options(responseType: ResponseType.json));
      // Get logged user
      User loggedUser = await _userDao.get();
      // Parse the new add address
      UserAddress parsedAddress = UserAddress.fromJson(response.data);
      // Update fields
      if (loggedUser.addresses == null) loggedUser.addresses = <UserAddress>[];
      loggedUser.addresses.add(parsedAddress);
      await _userDao.insertOrUpdate(loggedUser);
      // Callback invocation
      return ApiCallback<UserAddress>(isSuccessful: true, resultData: parsedAddress);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateUserAddress(UserAddress body) async {
    Map jsonBody = body.toJson();
    jsonBody.remove('id');
    jsonBody.remove('isDeleted');
    jsonBody.remove('landmark');
    try {
      await _legacyHttpRequester.patch(USER_ADDRESS.replaceAll("%", body.id), data: jsonBody, options: Options(responseType: ResponseType.json));
      // Load user
      await loadUser();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> deleteUserAddress(UserAddress body) async {
    try {
      await _legacyHttpRequester.delete(USER_ADDRESS.replaceAll("%", body.id), options: Options(responseType: ResponseType.json));
      // Load user
      await loadUser();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<PaymentMethod>>> loadUserPayments() async {
    try {
      var response = await _legacyHttpRequester.get(USER_PAYMENTS, options: Options(responseType: ResponseType.json));
      // Get logged user
      User loggedUser = await _userDao.get();
      // Fetch payment methods
      var responseData = response.data as List;
      var payments = List<PaymentMethod>();
      for (var row in responseData) {
        payments.add(PaymentMethod.fromJson(row));
      }
      // Update fields
      loggedUser.paymentMethods = payments;
      await _userDao.insertOrUpdate(loggedUser);
      // Callback invocation
      return ApiCallback<List<PaymentMethod>>(isSuccessful: true, resultData: payments);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<PaymentMethod>> createUserPayment(String number, String cardHolder, String cvv, String expiration, String brand, String cpfCardHolder,
      {PaymentMethodType type = PaymentMethodType.credit}) async {
    try {
      Map<String, String> body = {'number': number, 'cardHolder': cardHolder, 'cvv': cvv, 'expiration': expiration, 'brand': brand, 'cpfCardHolder': cpfCardHolder, 'type': type.index.toString()};
      var response = await _legacyHttpRequester.post(USER_PAYMENTS, data: body, options: Options(responseType: ResponseType.json));
      // Get logged user
      User loggedUser = await _userDao.get();
      // Parse the new add address
      PaymentMethod parsedPayment = PaymentMethod.fromJson(response.data);
      // Update fields
      if (loggedUser.paymentMethods == null) loggedUser.paymentMethods = <PaymentMethod>[];
      loggedUser.paymentMethods.add(parsedPayment);
      await _userDao.insertOrUpdate(loggedUser);
      // Callback invocation
      return ApiCallback<PaymentMethod>(isSuccessful: true, resultData: parsedPayment);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> deleteUserPayment(String paymentId) async {
    try {
      await _legacyHttpRequester.delete(USER_PAYMENT.replaceAll("%", paymentId), options: Options(responseType: ResponseType.json));
      // Load user
      await loadUser();
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<String>> loadCardBrand(String bin) async {
    try {
      var response = await _legacyHttpRequester.get(CARD_BRAND.replaceAll('%', bin), options: Options(responseType: ResponseType.json));

      CreditCardBrand brand = CreditCardBrand.fromJson(response.data);

      return ApiCallback(isSuccessful: true, resultData: brand.brandName);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> logoutUser() async {
    try {
      await _legacyHttpRequester.put(USER_LOGOUT, options: Options(responseType: ResponseType.json));
      // Remove user authorization
      await _authResponseDao.delete(AuthResponse.responseUser);
      // Remove the user locally
      await _userDao.deleteAll();
      // Force refresh on device token
      final token = await LoriPushNotificationToken.getStoredPushNotificationToken();
      await updateNotificationToken(token);
      // Callback invocation
      return ApiCallback<User>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// User part. All methods that are related with user.
  /// endregion [END]
  ///

  ///
  /// region Order part. All methods that are related with orders.
  /// [START]
  ///

  Future<ApiCallback<List<Order>>> loadOrders() async {
    try {
      var response = await _legacyHttpRequester.get(ORDERS, options: Options(responseType: ResponseType.json));
      var responseData = response.data as List;
      var orders = List<Order>();
      for (var row in responseData) {
        orders.add(Order.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Order>>(isSuccessful: true, resultData: orders);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<Order>> loadOrder(String orderId) async {
    try {
      var response = await _legacyHttpRequester.get(ORDER_BY_ID.replaceFirst('%', orderId), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback<Order>(isSuccessful: true, resultData: Order.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<PaymentMethod>>> loadOrderPayments(Order order) async {
    try {
      var response = await _legacyHttpRequester.get(ORDER_PAYMENTS.replaceFirst('%', order.id), options: Options(responseType: ResponseType.json));
      // Fetch payment methods
      var responseData = response.data as List;
      var payments = List<PaymentMethod>();
      for (var row in responseData) payments.add(PaymentMethod.fromJson(row));
      // Callback invocation
      return ApiCallback<List<PaymentMethod>>(isSuccessful: true, resultData: payments);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateOrderPayment(Order order, PaymentMethod paymentMethod) async {
    try {
      await _legacyHttpRequester.put(UPDATE_ORDER_PAYMENT.replaceFirst('%', order.id).replaceFirst('%', paymentMethod.id), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback<List<PaymentMethod>>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateOrderPaymentCvv(Order order, String cvv) async {
    try {
      await _legacyHttpRequester.put(UPDATE_ORDER_PAYMENT_CVV.replaceFirst("%", order.id).replaceFirst('%', cvv), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<List<Installment>>> loadOrderInstallments(Order order) async {
    try {
      final response = await _legacyHttpRequester.get(ORDER_INSTALLMENTS.replaceFirst('%', order.id), options: Options(responseType: ResponseType.json));
      // Parse result
      var responseData = response.data as List;
      var installments = List<Installment>();
      for (var row in responseData) {
        installments.add(Installment.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Installment>>(isSuccessful: true, resultData: installments);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> updateOrderInstallment(Order order, String installmentNumber) async {
    try {
      await _legacyHttpRequester.put(UPDATE_ORDER_INSTALLMENTS.replaceFirst("%", order.id).replaceFirst('%', installmentNumber), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> restartOrder(Order order) async {
    try {
      await _legacyHttpRequester.put(RESTART_ORDER.replaceFirst("%", order.id), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  Future<ApiCallback<void>> cancelOrder(String orderId) async {
    try {
      await _legacyHttpRequester.delete(ORDER_BY_ID.replaceFirst('%', orderId), options: Options(responseType: ResponseType.json));
      // Callback invocation
      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback(isSuccessful: false, resultErrorMessage: parseResponseError(e.response != null ? e.response.data : null));
    }
  }

  ///
  /// Order part. All methods that are related with orders.
  /// endregion [END]
  ///

  String parseResponseError(dynamic response) {
    if (response == null) {
      return null;
    } else if (response is Map<String, dynamic>) {
      Map<String, dynamic> map = response;
      if (map.containsKey('error')) {
        String errorDescription = map['error'];
        switch (errorDescription) {
          case 'invalid_grant':
            return 'Usuário e/ou senha inválido';
        }
      } else if (map.containsKey('message')) {
        return map['message'];
      } else if (map.containsKey('Message')) {
        return map['Message'];
      }
    }
    return null;
  }
}

class ApiCallback<T> {
  ApiCallback({this.isSuccessful, this.resultData, this.resultErrorMessage, this.error});

  bool isSuccessful;
  T resultData;
  String resultErrorMessage;
  DioError error;

  static String parseResponseError(dynamic response) {
    if (response == null) {
      return null;
    } else if (response is Map<String, dynamic>) {
      Map<String, dynamic> map = response;
      if (map.containsKey('error')) {
        String errorDescription = map['error'];
        switch (errorDescription) {
          case 'invalid_grant':
            return 'Usuário e/ou senha inválido';
        }
      } else if (map.containsKey('message')) {
        return map['message'];
      } else if (map.containsKey('Message')) {
        return map['Message'];
      }
    }
    return null;
  }

  ApiCallback<T> defaultErrorCallback(DioError e) {
    return ApiCallback<T>(isSuccessful: false, error: e, resultErrorMessage: ApiCallback.parseResponseError(e.response != null ? e.response.data : null));
  }
}
