import 'dart:async';

import 'package:lori_sdk/data/database/interfaces/searchTermsDao.dart';

class SearchedTermsSingleton {
  // Singleton part
  static final SearchedTermsSingleton _singleton =
      SearchedTermsSingleton._internal();

  factory SearchedTermsSingleton() {
    return _singleton;
  }

  SearchedTermsSingleton._internal();

  // Class part
  // ignore: close_sinks
  var _searchTerms = StreamController<List<String>>.broadcast(sync: true);

  Future<StreamSubscription<List<String>>> registerToStream(
      Function callback) async {
    StreamSubscription<List<String>> subscription =
        _searchTerms.stream.listen(callback);
    // Make sure to call cart changed this time
    var terms = await SearchTermsDao().get();
    searchTermsChanged(terms);
    return subscription;
  }

  void searchTermsChanged(List<String> terms) {
    _searchTerms.sink.add(terms);
  }

  void userClose() async {
    await _searchTerms.sink.close();
  }
}
